// Copyright (C) 2015, Antoni Buades, Universitat Illes Balears
// // <toni.buades@uib.es>
// //
// //
// //
// // This program is free software: you can use, modify and/or
// // redistribute it under the terms of the GNU General Public
// // License as published by the Free Software Foundation, either
// // version 3 of the License, or (at your option) any later
// // version. You should have received a copy of this license along
// // this program. If not, see <http://www.gnu.org/licenses/>.
//
//
//


#include <string.h>
#include <sstream>
#include "libCartoon.h"


namespace libUSTG
{

/*******************************************************************************
Cartoon+Texture decomposition
Input:
* input image
* scale parameter (fSigma)
* angular step of directional filters (fAngle)
* flag indicating use of isotropic filter (iflag2D)
Output:
* output image (cartoon)
 
*******************************************************************************/
  
void non_linear_cartoon_newdirectional(cflimage &input, cflimage &out,
                                       float fSigma, float fAngle,
                                       int iflag2D)
{

    //! kernel stuff
    // number of directions
    int iDirections = (int) rintf(180 / fAngle);
    iDirections*=2; //two polarities per sign
    if (iflag2D) iDirections++;

    // exponential decay in backward direction
    float fXsigma = 0.75;
    // exponential decay in forward direction
    float fYsigma = fSigma;
    int ksize = 4 * (int) ceilf(MAX(fXsigma, fYsigma)) + 1;

    //! Gradient computation
    cflimage grad = input.gradient('c');

    //memory allocation for the optimum low pass filtered value for each pixel
    cflimage wta(input.w(), input.h(), input.c());
  
    //memory allocation for the maximum
    //local total variation rate at each pixel
    flimage wtaW(input.w(), input.h());
  
    //initialization
    wta = input;
    wtaW = 0.0f;

    float pkernel=0.99f; //parameter for filter kernel definition
    //(needed in sort_kernel() function)

    //memory allocation for the local total variation rates at each pixel
    float **diffsDir=new float*[iDirections];
    for (int iI=0; iI < iDirections; iI++)
      diffsDir[iI]=new float[input.w()*input.h()];
  
    //memory allocation for the optimal directions of filtering of each pixel
    int *dir=new int[input.w()*input.h()];

    //memory allocation for the set of low filtered images
    cflimage *low=new cflimage[iDirections];

    //! Process
    int sign=1;
    #pragma omp parallel for schedule(dynamic)
    for (int iI = 0; iI < iDirections; iI++) {

        //needed inside the for loop when using paralelization
        libUSTG::flimage kernel(ksize, ksize);
        //memory allocation for filter kernel
        struct sorted_kernel *skernel=new struct sorted_kernel;
        skernel->nitems=ksize*ksize;
        skernel->w=ksize;
        skernel->h=ksize;
        skernel->ncut=0;
        skernel->items=new kernel_item[skernel->nitems];

        //! Create oriented kernel
        float fiAngle = fAngle * (float) (iI/2);

        //printf("direction %i, angle %2.2f\n", iI, fiAngle);

        //Create filter kernel
        //if iflagD active, add isotropic kernel to the list of  filters
        if (iflag2D && iI == iDirections-1)
            fiFloatDirectionalGaussKernel(fYsigma,fYsigma, 0.0, kernel.v(0), \
                                          kernel.w(), kernel.h());
        else
            fiFloatDirectionalGaussKernelS(fXsigma,fYsigma, fiAngle, \
                                           kernel.v(0), kernel.w(), kernel.h(),\
                                           sign);

        // ! To speed up computations:
        //  Organize kernel values in decreasing order, keep the index of the
        //  position with pkernel % of total area
        sort_kernel(kernel.v(), kernel.w(), kernel.h(), skernel, pkernel);
        //printf("sorted kernel %i: %i/%i\n",iI,skernel->ncut, skernel->nitems);

        //! Low pass filtering
        low[iI] = input;
        int niter = 5;
        for (int iC = 0; iC < input.c(); iC++)
            low_pass_filter_skernel(input.v(iC), low[iI].v(iC), skernel, niter,
                                    input.w(), input.h());

        //! ratio1 <- G * |Df|
        cflimage gradC = input;
        grad.convolve_skernel(skernel, BOUNDARY_CONDITION_SYMMETRIC, &gradC);

        //! add values to get ratio1
        flimage ratio1(input.w(), input.h());
        ratio1 = gradC.getGray(0.33, 0.33, 0.33);

        //!  ratio2 <- G*|D(L_sigma*f)|
        cflimage lowG = low[iI].gradient('c');
        cflimage lowGC = lowG;
        lowG.convolve_skernel(skernel, BOUNDARY_CONDITION_SYMMETRIC, &lowGC);

        //! add values to get ratio2
        flimage ratio2(input.w(), input.h());
        ratio2 = lowGC.getGray(0.33, 0.33, 0.33);


        //compute local total variation rates at each pixel
        for (int iPix = 0; iPix < input.wh(); iPix++) {

            float difference = ratio1[iPix] - ratio2[iPix];
            float ar1 = (float) fabs((double) ratio1[iPix]);

            if (ar1 > 0.0f)
                difference /= ar1;
            else
                difference = 0.0;

            diffsDir[iI][iPix]=difference;

        }

        //free memory
        //needed inside the for loop when using paralelization
        delete[] skernel->items;
        delete skernel;

        //reverse direction
        sign*=-1;

    }

    //assign direction of maximum variation to each pixel
    for (int iI = 0; iI < iDirections; iI++) {

        for (int iPix = 0; iPix < input.wh(); iPix++) {

            if (iI == 0) {//first value
                wtaW[iPix] = diffsDir[iI][iPix];
                dir[iPix] = iI;
            } else {
                if (diffsDir[iI][iPix] > wtaW[iPix]) { //take max
                    wtaW[iPix] = diffsDir[iI][iPix];
                    dir[iPix] = iI;
                }
            }
        }
    }


    //assign optimum low-pass filtered value to each pixel
    for (int iI = 0; iI < iDirections; iI++) {
        for (int iPix = 0; iPix < input.wh(); iPix++) {
            if (dir[iPix] == iI) {
                for (int iC=0; iC < input.c(); iC++)
                    wta[iC * input.wh() + iPix] = low[iI][iC * input.wh()+iPix];
            }
        }
    }
  
    //get output image (cartoon)
    for (int iPix = 0; iPix < input.wh(); iPix++)
    {
      //define weighting function
      float weight;
      float cmin= 0.25f;
      float cmax= 0.50f;
      
      if (wtaW[iPix] < cmin)
      weight = 0.0;
      else if (wtaW[iPix] > cmax)
      weight = 1.0;
      else
      weight = (wtaW[iPix] - cmin) / (cmax - cmin);
      
      //combine original and low-pass filtered values
      for (int iC=0; iC < input.c(); iC++)
      {
        out[iC * input.wh() + iPix] = weight * wta[iC * input.wh() + iPix] + \
                                  (1.0f-weight) * input[iC * input.wh() + iPix];
      }
      
    }
  
  
    //clean up
    delete[] dir;
    for (int iI = 0; iI < iDirections; iI++) delete[] diffsDir[iI];
    delete[] diffsDir;
    delete[] low;

}




/**
 * \brief   Low Pass filter
 *  The low pass filter is approximated by   L = Id - (Id - K)^{n *}
 *  That is, the difference between f and K(f) is convolved again with
 *  K, and the result of this convolution is put back to f
 *  This approach is iterated niter times
 *
 *
 * @param[in]   input   input image
 * @param[out]  out             output vector
 * @param[in]   skernel    filter kernel
 * @param[in]   niter   number of iterations
 * @param[in]   width, height           size of the image
 *
 */


void low_pass_filter_skernel(float * input, float * out,
                             struct sorted_kernel *skernel, int niter,
                             int width, int height)
{


    int size = width*height;


    // memory for K * f   and   f - K * f
    float *imconvolved = (float *) malloc(size*sizeof(float));
    float *imdifference = (float *) malloc(size*sizeof(float));


    // out  <-  K * f
    fiConvol_skernel(input, out, width, height, skernel, \
                     BOUNDARY_CONDITION_SYMMETRIC);

    if (niter > 0) {

        // imdifference <-  input - out
        fpCombine(input,1.0,out,-1.0,imdifference,width*height);


        for(int i=0; i<niter; i++) {

            // imconvolved <-  K * imdifference

            fiConvol_skernel(imdifference, imconvolved, width, height, \
                             skernel, BOUNDARY_CONDITION_SYMMETRIC);
            // imdifference <- imdifference - K * imdifference
            fpCombine(imdifference,1.0,imconvolved,-1.0,
                      imdifference,width*height);

        }


        // out <- input - imdifference
        fpCombine(input,1.0,imdifference,-1.0,out,size);


    }


    if (imdifference) free(imdifference);
    if (imconvolved) free(imconvolved);

}







}

