// Copyright (C) 2015, Antoni Buades, Universitat Illes Balears
// // <toni.buades@uib.es>
// //
// //
// //
// // This program is free software: you can use, modify and/or
// // redistribute it under the terms of the GNU General Public
// // License as published by the Free Software Foundation, either
// // version 3 of the License, or (at your option) any later
// // version. You should have received a copy of this license along
// // this program. If not, see <http://www.gnu.org/licenses/>.
//
//
//


#ifndef __libImage__
#define __libImage__



extern "C" {
#include "libImageFormats.h"
}


#define USE_FFTW
#ifdef USE_FFTW
#include <fftw3.h>
#endif

#include "libImageFormatPM.h"
#include "libBasic.h"


#define iipHorizontal 0
#define iipVertical 1



namespace libUSTG
{


//
//! Class definitions
//

class cflimage;
class flimage;


//////////////////////////////////////////////////////////////////////////
//
//! Image Classes
//
////////////////////////////////////////////////////////////////////////


class cflimage
{


protected:



public:

    int d_c, d_w, d_h, d_wh, d_whc;		// channels, width, height,
                                      // width*height, width*height*channels
    float *d_v;							// pointer


    char name[128];
    float visuMin, visuMax;				// useful only for visualization with mxview



    //
    //! Construction / Memory
    //

    cflimage();
    cflimage(int w, int h, int c);
    cflimage(int w, int h, float *igray);
    cflimage(int w, int h, float *ired, float *igreen, float *iblue);

    cflimage(int w, int h, unsigned char *igray);
    cflimage(int w, int h, unsigned char *ired, unsigned char *igreen,
             unsigned char *iblue);

    cflimage(const flimage &red, const flimage &green, const flimage &blue);


    void create(int w, int h, int c);

    cflimage(const cflimage& im);

    void erase();
    virtual ~cflimage();



    //
    //! Operators
    //

    cflimage& operator= (const cflimage& im);



    //
    //! Load / Save
    //
    void load(const char *filename);
    void save(const char *filename);



    //
    //! Get Basic Data
    //

    int c() const {
        return d_c;
    }
    int w() const {
        return d_w;
    }
    int h() const {
        return d_h;
    }
    int wh() const {
        return d_wh;
    }
    int whc() const {
        return d_whc;
    }


    float * v() {
        return d_v;
    }
    float * v(int i) {
        /*assert(i>= 0 && i < d_c);*/ return &d_v[i * d_wh];
    }

    inline float operator[](int i) const {
        /*assert(i>=0 && i < d_whc);*/ return d_v[i];
    }
    inline float& operator[](int i) {
        /*assert(i>=0 && i < d_whc);*/ return d_v[i];
    }


    flimage getChannel(int i);
    void getChannel(int i, flimage *out);

    operator  flimage();



    int isSameSize(const  cflimage &inIm);

    char * getName() {
        return name;
    }
    void  setName(const char *iname) {
        strcpy(name, iname);
    }


    float getVisuMin() {
        return visuMin;
    }
    float getVisuMax() {
        return visuMax;
    }

    void setVisuMin(float a) {
        visuMin = a;
    }
    void setVisuMax(float a) {
        visuMax = a;
    }


    //
    //! Math
    //

    cflimage& operator= (float a);
    void operator-=(float a);
    void operator+=(float a);
    void operator*=(float a);

    float  min();
    float  max();

    float  min_channel (int i);
    float  max_channel (int i);

    void  max (float M);
    void  min (float m);

    void normalizeL1();

    void rint();
    void abs();

    void thre(float m, float M);



    //
    //! Color Conversion
    //

    flimage getGray();
    flimage getGray(float wr, float wg, float wb);

    void getGray(flimage * out);
    void getGray(float wr, float wg, float wb, flimage *out);


    //
    //! Gradient
    //
    cflimage gradient(char cType);
    cflimage xgradient(char cType);
    cflimage ygradient(char cType);
    cflimage gradient(cflimage &orientation, char cType);
    cflimage gradient(cflimage &xgrad, cflimage &ygrad, cflimage &orientation,
                      char cType);



    //
    //! Subsampling & zooming & convolution
    //
    void convolve_skernel(struct sorted_kernel *skernel, int boundary,
                          cflimage *out);



    //
    //! Value operations
    //
    void addGaussianNoise(float std);
    void addGaussianNoiseSD(float std, float sdstd);



};



class flimage : public cflimage
{

public:


    flimage();
    flimage(int w, int h);
    flimage(int w, int h, float *ptr);
    flimage(int w, int h, unsigned char *ptr);

    flimage(const flimage& im);


    void create(int w, int h);


    using cflimage::operator=;
    flimage& operator=(const flimage& im);


    float operator()(int i, int j) const {
        return this->d_v[j * this->d_w + i];
    }
    float& operator()(int i, int j) {
        return this->d_v[j * this->d_w + i];
    }



    virtual ~flimage() {};

};







}





#endif
